"""Util function for backend."""
from typing import Dict, Callable

import numpy as np
from braket.circuits import Instruction, Circuit, result_types, gates
from numpy import pi

import qutrunk.circuit.gates as i_gates
from .exception import QuTrunkBraketException


def u3(theta, phi, lam):
    return [
        gates.Rz(lam),
        gates.Rx(pi / 2),
        gates.Rz(theta),
        gates.Rx(-pi / 2),
        gates.Rz(phi),
    ]


qutrunk_gate_names_to_braket_gates: Dict[str, Callable] = {
    "u1": lambda lam: [gates.Rz(lam)],
    "u2": lambda phi, lam: [gates.Rz(lam), gates.Ry(pi / 2), gates.Rz(phi)],
    "u3": lambda theta, phi, lam: u3(theta, phi, lam),
    "p": lambda angle: [gates.PhaseShift(angle)],
    "cp": lambda angle: [gates.CPhaseShift(angle)],
    "cx": lambda: [gates.CNot()],
    "x": lambda: [gates.X()],
    "y": lambda: [gates.Y()],
    "z": lambda: [gates.Z()],
    "t": lambda: [gates.T()],
    "tdg": lambda: [gates.Ti()],
    "s": lambda: [gates.S()],
    "sdg": lambda: [gates.Si()],
    "sx": lambda: [gates.V()],
    "sxdg": lambda: [gates.Vi()],
    "swap": lambda: [gates.Swap()],
    "r": lambda theta, phi: u3(theta, phi - pi / 2, -phi + pi / 2),
    "rx": lambda angle: [gates.Rx(angle)],
    "ry": lambda angle: [gates.Ry(angle)],
    "rz": lambda angle: [gates.Rz(angle)],
    "rzz": lambda angle: [gates.ZZ(angle)],
    "id": lambda: [gates.I()],
    "h": lambda: [gates.H()],
    "cy": lambda: [gates.CY()],
    "cz": lambda: [gates.CZ()],
    "ccx": lambda: [gates.CCNot()],
    "cswap": lambda: [gates.CSwap()],
    "rxx": lambda angle: [gates.XX(angle)],
    "ryy": lambda angle: [gates.YY(angle)],
    "ecr": lambda: [gates.ECR()],

    "x1": lambda: [gates.Rx(pi / 2)],
    "y1": lambda: [gates.Ry(pi / 2)],
    "z1": lambda: [gates.Rz(pi / 2)],
    "iswap": lambda: [gates.Unitary(np.array(i_gates.iSwap.matrix), "iSwap")],
    "sqrtx": lambda: [gates.V()],
    "cr": lambda theta: [gates.Unitary(np.array(i_gates.CR(theta).matrix), "CR")],
    "crx": lambda theta: [gates.Unitary(np.array(i_gates.CRx(theta).matrix), "CRx")],
    "cry": lambda theta: [gates.Unitary(np.array(i_gates.CRy(theta).matrix), "CRy")],
    "crz": lambda theta: [gates.Unitary(np.array(i_gates.CRz(theta).matrix), "CRz")],
    "cu": lambda theta, phi, lam, gamma: [gates.Unitary(np.array(i_gates.CU(theta, phi, lam, gamma).matrix), "CU")],
    "cu1": lambda theta: [gates.Unitary(np.array(i_gates.CU1(theta).matrix), "CU1")],
    "cu3": lambda theta, phi, lam: [gates.Unitary(np.array(i_gates.CU3(theta, phi, lam).matrix), "CU3")],
    "ch": lambda: [gates.Unitary(np.array(i_gates.CH.matrix), "CH")],
    "csqrtx": lambda: [gates.CV()],
    "sqrtxdg": lambda: [gates.Vi()],
}


def convert_qutrunk_to_braket_circuit(circuit) -> Circuit:
    """Return a Braket quantum circuit from a QuTrunk quantum circuit.
     Args:
            circuit (QCircuit): QuTrunk Quantum Cricuit

    Returns:
        Circuit: Braket circuit
    """
    quantum_circuit = Circuit()
    for qutrunk_cmd in circuit.cmds:
        name = str(qutrunk_cmd.gate).lower()
        if name == "measure":
            # TODO: change Probability result type for Sample for proper functioning # pylint:disable=fixme
            quantum_circuit.add_result_type(
                result_types.Probability(
                    target=[
                        qutrunk_cmd.targets[0],
                        qutrunk_cmd.targets[0]
                    ]
                )
            )
        elif name == "barrier":
            pass
        else:
            params = []
            if qutrunk_cmd.rotation:
                params = qutrunk_cmd.rotation

            ctr_cnt = len(qutrunk_cmd.controls)
            if name == "mcx":
                if ctr_cnt == 1:
                    gate_list = [gates.CNot()]
                elif ctr_cnt == 2:
                    gate_list = [gates.CCNot()]
                else:
                    raise QuTrunkBraketException(
                        f"{name}({ctr_cnt}):mcx with more than 2 control bit isn't supported by BRAKET")
            elif name == "mcz":
                if ctr_cnt == 1:
                    gate_list = [gates.CZ()]
                else:
                    raise QuTrunkBraketException(
                        f"{name}({ctr_cnt}):mcz with more than 1 control bit isn't supported by BRAKET")
            else:
                gate_list = qutrunk_gate_names_to_braket_gates[name](*params)

            if not gate_list:
                raise QuTrunkBraketException(f"{name} isn't supported by BRAKET")

            for gate in gate_list:
                instruction = Instruction(
                    # Getting the index from the bit mapping
                    operator=gate,
                    target=qutrunk_cmd.controls + qutrunk_cmd.targets,
                )
                quantum_circuit += instruction
    return quantum_circuit



"""Circuit Module."""
import json
import random
from typing import List, Optional, Union, Callable
import numpy as np

from qutrunk.backends import Backend, BackendLocal
from qutrunk.circuit import CBit, CReg, Counter, QuBit, Qureg
from qutrunk.circuit.gates import BarrierGate, MeasureGate, Observable, PauliCoeffs
from qutrunk.circuit.parameter import Parameter
from qutrunk.circuit.ops import AMP
from qutrunk.exceptions import QuTrunkError


class QCircuit:
    """Quantum circuit.

    Args:
        backend: Used to run quantum circuits.
        density: Creates a density matrix Qureg object representing a set of \
            qubits which can enter noisy and mixed states.
        name: The circuit name.
        resource: Whether enable the resource statistics function, default: False.

    Example: 
        .. code-block:: python

            from qutrunk.circuit import QCircuit
            from qutrunk.circuit.gates import H, CNOT, Measure

            # new QCircuit object
            qc = QCircuit()
            qr = qc.allocate(2)
            H * qr[0]
            CNOT * (qr[0], qr[1])
            Measure * qr[0]
            Measure * qr[1]
            res = qc.run(shots=100)
    """

    prefix = "circuit"

    def __init__(
        self,
        backend=None,
        density=False,
        name: Optional[str] = None,
        resource: Optional[bool] = False,
    ):
        self.qreg = None
        self.creg = None
        self.cmds = []
        self.cmd_cursor = 0
        self.counter = None
        # mark circuit in Operator Context
        self._in_op = False

        self.qubit_indices = {}
        self.cbit_indices = {}

        # 参数字典{Parameter: value}
        self.param_dict = {}

        # use local backend(default)
        if backend is None:
            self.backend = BackendLocal()
        else:
            if not isinstance(backend, Backend):
                raise TypeError("You supplied a backend which is not supported.\n")
            self.backend = backend

        # local backend is no support noisy pattern.
        if density and isinstance(backend, BackendLocal):
            raise TypeError("You supplied a backend which is not supported density.\n")
        self.density = density

        self.backend.circuit = self
        # TODO:?
        self.outcome = None

        if name is None:
            name = self._generate_circuit_name()
        self.name = name

        if resource:
            self.counter = Counter(self)
    
    def allocate(self, qubits: Union[int, list]):
        """Allocate qubit in quantum circuit.

        Args:
            qubits: int: The number of qubit allocated in circuit.\
                    list: The sum of list is the number of qubit allocated in circuit,\
                    and each value item represents the size of corresponding subqureg.

        Returns:
            The register of quantum.
        """

        if not isinstance(qubits, (int, list)):
            raise TypeError("qubits parameter should be type of int or list.")

        qubit_size = qubits if isinstance(qubits, int) else sum(qubits)
        if qubit_size <= 0:
            raise TypeError("Number of qubits should be larger than 0.")

        if qubit_size > 20:
            raise ValueError("Number of qubits should be less than 20.")

        self.qreg = Qureg(circuit=self, size=qubit_size)
        self.creg = CReg(circuit=self, size=qubit_size)

        if self.counter:
            self.counter.qubits = qubit_size

        for index in range(qubit_size):
            self.qubit_indices[QuBit(self.qreg, index)] = index
            self.cbit_indices[CBit(self.creg, index)] = index

        if isinstance(qubits, int):
            return self.qreg
        elif isinstance(qubits, list):
            return self.qreg.split(qubits)

    def set_cmds(self, cmds):
        """Set cmds to circuit.

        Args:
            cmds: The commands set to circuit.
        """
        self.cmds = cmds

    def append_cmd(self, cmd):
        """Append command to circuit when apply a quantum gate.

        Args:
            cmd: The command append to circuit.
        """
        self.cmds.append(cmd)

    def append_circuit(self, circuit):
        """Append target circuit to current circuit.

        Note: The target circuit must have the same qubits as current circuit.

        Args:
            circuit: The target circuit append to current circuit.
        """
        if circuit.num_qubits != self.num_qubits:
            raise QuTrunkError("The target circuit must have the same qubits as current circuit.")
        for cmd in circuit.cmds:
            self.append_cmd(cmd)

    # TODO: need to improve.
    def forward(self, num):
        """Update the cmd_cursor when a bunch of quantum operations have been run.

        Args:
            num: The number of cmds in current bunch.
        """
        self.cmd_cursor += num

    @property
    def num_qubits(self) -> int:
        """Get the number of qubits.

        Returns:
            The length of the self.qreg.
        """

        return len(self.qreg)

    @property
    def num_gates(self) -> int:
        """Get the number of gates.

        Returns:
            The length of the self.cmds.
        """
        return len(self.cmds)

    def set_measure(self, qubit, value):
        """Store the measure result for target qubit.

        Args:
            qubit: The index of qubit in qureg.
            value: The qubit measure value(0 or 1).

        Raises:
            IndexError: Qubit index must be less than then length of qreg.
        """
        if qubit >= len(self.qreg) or qubit < 0:
            raise IndexError("qubit index out of range.")
        self.creg[qubit].value = value

    # TODO: need to improve.
    def run(self, shots=1):
        """Run quantum circuit through the specified backend and shots.

        Args:
            shots: Run times of the circuit, default: 1.

        Returns:
            The Result object contain circuit running outcome.
        """
        self.backend.send_circuit(self, True)
        result = self.backend.run(shots)

        if self.backend.name == "BackendIBM":
            # note: ibm后端运行结果和qutrunk差异较大，目前直接将结果返回不做适配
            return result
        # TODO: measureSet
        if result and result.measureSet:
            for m in result.measureSet:
                self.set_measure(m.id, m.value)

        res = Result(self.num_qubits, result, self.backend, arguments={"shots": shots})

        return res

    def draw(self, output=None, line_length=None):
        """Quantum circuit.

        Draw the quantum circuit.

        Args:
            output: Set the method of drawing the circuit.
            line_length: Max length to draw quantum circuit, the excess circuit will be truncated.

        Returns:
            Visualized quantum circuit diagram.
        """
        if output is None:
            output = "text"

        from qutrunk.visualizations import circuit_drawer

        print(circuit_drawer(circuit=self, output=output, line_length=line_length))

    def __str__(self) -> str:
        return str(self.draw(output="text"))

    @property
    def qubits(self) -> List[QuBit]:
        """Returns a list of quantum bits."""
        return self.qreg.qubits

    @property
    def cbits(self) -> List[CBit]:
        """Returns a list of cbit."""
        return self.creg.cbits

    def _generate_circuit_name(self):
        """Generate circuit name."""
        return f"{self.prefix}-{str(random.getrandbits(15))}"

    def __len__(self) -> int:
        """Return the number of operations in circuit."""
        return len(self.cmds)

    def get_prob(self, value):
        """Get probability of the possible measure result of circuit.

        Args:
            value: The target value.

        Returns:
            float: The probability of value.
        """
        self.backend.send_circuit(self)
        return self.backend.get_prob(value)

    def get_probs(self):
        """Get all probabilities of circuit.

        Returns:
            An array contains all probabilities of circuit.
        """
        qubits = [i for i in range(self.num_qubits)]
        self.backend.send_circuit(self)
        probs = self.backend.get_probs(qubits)
        
        out_probs = []
        for i in range(len(probs)):
            prob = {}
            prob["idx"] = i
            prob["prob"] = probs[i]
            out_probs.append(prob)
        return out_probs

    def _to_complex(self, state_vector):
        result = []
        for item in state_vector:
            r = item.split(",")
            result.append(complex(float(r[0]), float(r[1])))
        return result

    def get_statevector(self):
        """Get state vector of circuit."""
        self.backend.send_circuit(self)
        return self._to_complex(self.backend.get_statevector())

    def find_bit(self, bit):
        """Find locations in the circuit.

        Returns the index of the qubit or CBit in the circuit.

        Args:
            bit: QuBit or CBit.

        Returns:
            The index of QuBit or CBit in circuit.
        """
        try:
            if isinstance(bit, QuBit):
                return self.qubit_indices[bit]
            elif isinstance(bit, CBit):
                return self.cbit_indices[bit]
            else:
                raise Exception(f"Could not locate bit of unknown type:{type(bit)}")
        except KeyError:
            raise Exception(f"Could not locate provided bit:{bit}")

    def create_parameter(self, name: str):
        """
        Allocate one Parameter object.

        Args:
            name(str): Parameter name.

        Returns:
            Parameter object
        """
        p = Parameter(name)
        self.param_dict[name] = p
        return p

    def create_parameters(self, names: list)->tuple:
        """
        Allocate a batch of parameters.

        Args:
            names(list): A list of parameter's name.

        Returns:
            tuple: a batch of parameters. 
        """ 
        params = []
        for name in names:
            params.append(self.create_parameter(name))

        return tuple(params)

    def get_parameters(self):
        """Get all parameters of circuit.

        Args:
            list: all parameters in circuit.
        """
        param_values = [param.value for param in self.param_dict.values()]
        return param_values

    def bind_parameters(self, params):
        """
        Assign specific value to parameters.

        Args:
            params (dict): {parameter: value, ...}.

        Raises:
            ValueError: parameters variable contains parameters not present in the circuit.
        """
        if not isinstance(params, dict):
            raise ValueError("parameters must be dictionary.")
        # 1 参数是否在参数表�
        parameters_table_key = self.param_dict.keys()
        params_not_in_circuit = [
            param_key
            for param_key in params.keys()
            if param_key not in parameters_table_key
        ]
        if len(params_not_in_circuit) > 0:
            raise ValueError(
                f"Cannot bind parameters ({', '.join(map(str, params_not_in_circuit))}) "
                f"not present in the circuit."
            )

        # update parameter
        for k, v in params.items():
            param = self.param_dict[k]
            param.update(v)
            

        # note: 绑定参数后意味着线路已经改变，需要重新构建线路
        new_circuit = QCircuit(backend=self.backend, name=self.name)
        new_circuit.allocate(qubits=self.num_qubits)
        new_circuit.set_cmds(self.cmds)
        return new_circuit

    def inverse(self):
        """Invert this circuit.

        Reverses the circuit and returns an error message.

        Returns:
            tuple: (QCircuit,qreg)

        Raises:
            ValueError: if the circuit cannot be inverted.
        """
        inverse_circuit = QCircuit(backend=self.backend, name=self.name + "_dg")
        inverse_circuit.allocate(qubits=self.num_qubits)

        # inverse cmd and gate
        cmds = self.cmds
        for cmd in reversed(cmds):
            if isinstance(cmd.gate, (MeasureGate, AMP)):
                raise ValueError("The circuit cannot be inverted.")
            cmd.inverse = True
            inverse_circuit.append_cmd(cmd)

        return inverse_circuit, inverse_circuit.qreg

    @staticmethod
    def load(file, format=None):
        """Deserialize file object containing a OpenQASM or qusl document to a Python object.

        Args:
            file (str): Path to the file for a qusl or OpenQASM program.
            format(str): The format of file content.

        Return:
            The QCircuit object for the input qusl or OpenQASM.

        """
        if format is None or format == "qusl":
            from qutrunk.tools.qusl_parse import qusl_to_circuit

            return qusl_to_circuit(file)

        if format == "openqasm":
            from qutrunk.qasm import Qasm
            from qutrunk.converters import dag_to_circuit
            from qutrunk.converters import ast_to_dag

            qasm = Qasm(file)
            ast = qasm.parse()
            dag = ast_to_dag(ast)
            return dag_to_circuit(dag)

    def expval_pauli(self, paulis: Union[list, object]):
        """Computes the expected value of a product of Pauli operators.

        Args:
            paulis:
                oper_type (int): Pauli operators.
                target (int): indices of the target qubits.

        Returns:
            The expected value of a product of Pauli operators.
        """
        pauli_list = []
        if not isinstance(paulis, list):
            pauli_list.append(paulis)
        else:
            pauli_list = paulis 
        self.backend.send_circuit(self)
        expect = self.backend.get_expec_pauli_prod(pauli_list)
        return expect

    def expval_pauli_sum(self, pauli_coeffs: PauliCoeffs):
        """Computes the expected value of a sum of products of Pauli operators.

        Args:
            pauli_coeffs (PauliCoeffs): Maintain a list of PauliCoeff, each PauliCoeff consist of one coefficient \
                and a series of pauli operators, PauliCoeffs is used to calculate the sum of Pauli products.

         .. note::

            The length of paulis in each term should be equal to self.num_qubits, otherwise \
                PauliI operator will be filled automatically.

        Returns:
            Returns the sum of Pauli products.

        Raises:
            ValueError: If the length of paulis in each term greater than self.num_qubits.
        """
        self.backend.send_circuit(self)
        paulis = []
        coeffs = []
        for term in pauli_coeffs:
            if len(term.paulis) > self.num_qubits:
                raise ValueError("The length of paulis in each term should be equal to self.num_qubits")
            if len(term.paulis) < self.num_qubits:
                term.padding(self.num_qubits - len(term.paulis))
            coeffs.append(term.coeff)
            paulis.extend(term.paulis)

        return self.backend.get_expec_pauli_sum(paulis, coeffs)

    def _dump_qusl(self, file):
        with open(file, "w", encoding="utf-8") as f:
            qusl_data = {}
            qusl_data["target"] = "QuSL"
            qusl_data["version"] = "1.0"

            meta = {"circuit_name": self.name, "qubits": str(len(self.qreg))}
            qusl_data["meta"] = meta

            inst = []
            for c in self.cmds:
                inst.append(c.qusl() + "\n")

            qusl_data["code"] = inst
            f.write(json.dumps(qusl_data))

    def _dump_openqasm(self, file):
        with open(file, "w", encoding="utf-8") as f:
            f.write("OPENQASM 2.0;\n")
            f.write('include "qelib1.inc";\n')
            f.write(f"qreg q[{str(len(self.qreg))}];\n")
            f.write(f"creg c[{str(len(self.qreg))}];\n")
            for c in self.cmds:
                f.write(c.qasm() + ";\n")

    def dump(self, file=None, format=None):
        """Serialize Quantum circuit as a JSON formatted stream to file.

        Args:
            file: Dump the qutrunk instruction to file(json format).
        """
        if file is None:
            raise Exception("file argument need to be supplied.")

        if format is None or format == "qusl":
            self._dump_qusl(file)

        if format == "openqasm":
            self._dump_openqasm(file)

    def _print_qusl(self):
        """Print quantum circuit in qutrunk form."""
        print(f"qreg q[{str(len(self.qreg))}]")
        print(f"creg c[{str(len(self.qreg))}]")

        for c in self.cmds:
            print(c.qusl())

    def _print_qasm(self):
        """Print quantum circuit in OpenQASM form."""
        print("OPENQASM 2.0;")
        print('include "qulib1.inc";')
        print(f"qreg q[{str(len(self.qreg))}];")
        print(f"creg c[{str(len(self.qreg))}];")
        for c in self.cmds:
            print(c.qasm() + ";")

    def print(self, format=None):
        """Print quantum circuit.

        Args:
            format(str): The format of needed to print, Default to qusl. options are "qusl" or "openqasm".
        """
        if format is None or format == "qusl":
            self._print_qusl()

        if format == "openqasm":
            self._print_qasm()

    # TODO: need to improve.
    def depth(
        self,
        counted_gate: Optional[Callable] = lambda x: not isinstance(x, BarrierGate),
    ) -> int:
        """Return circuit depth (i.e., max length of critical path).

        Args:
            counted_gate (callable): A function to filter out some instructions.
                Should take as input a tuple of (Instruction, list(Qubit), list(CBit)).
                By default filters out barrier.

        Returns:
            int: Depth of circuit.

        """

        # If no qubits or cmds, return 0
        if not self.qreg or not self.cmds:
            return 0

        # A list that holds the depth of each qubit
        depths = [0] * len(self.qreg)

        for cmd in self.cmds:
            if not counted_gate(cmd):
                continue

            indices = cmd.controls + cmd.targets
            cur_depths = []
            for idx1 in indices:
                cur_depths.append(depths[idx1] + 1)

            max_depth = max(cur_depths)
            for idx2 in indices:
                depths[idx2] = max_depth

        return max(depths)

    def width(self) -> int:
        """Get width of circuit, namely number of QuBit plus CBit.

        Returns:
            int: Width of circuit.

        """
        return len(self.qreg) + len(self.creg)

    def show_resource(self):
        """Show resource of the circuit use."""
        if self.counter:
            self.counter.show_verbose()


class Result:
    """Result of quantum operation.

    Save the result of quantum circuit running.

    Args:
        num_qubits: The number of qubits.
        res: The circuit running result from backend.
        backend: The backend that supports the operation of quantum circuits.
        task_id: Task id will automatic generate when submit a quantum computing job.
        status: The operating state of a quantum circuit.

    Example:
        .. code-block:: python

            from qutrunk.circuit import QCircuit
            from qutrunk.circuit.gates import H, Measure

            # new QCircuit object
            qc = QCircuit()
            qr = qc.allocate(1)
            H * qr[0]
            Measure * qr[0]
            # get running result
            res = qc.run()
            # get measurement from result
            print(res.get_measure())
    """

    def __init__(
        self, num_qubits, res, backend, arguments, task_id=None, status="success"
    ):
        self.states = []
        self.values = []
        self.backend = backend
        self.task_id = task_id
        self.status = status
        self.arguments = arguments
        # Modified quantum gate annotation.
        self.measure_result = [-1] * num_qubits

        if res and res.measureSet:
            for m in res.measureSet:
                self.set_measure(m.id, m.value)

        # Count the number of occurrences of the bit-bit string composed of all qubits.
        if res and res.outcomeSet:
            self.outcome = res.outcomeSet
            for out in self.outcome:
                # 二进制字符串转换成十进制
                if out.bitstr:
                    self.states.append(int(out.bitstr, base=2))
                    self.values.append(out.count)
        else:
            self.outcome = None

    def set_measure(self, qubit, value):
        """Update qubit measure value.

        Args:
            qubit: The index of qubit in qureg.
            value: The qubit measure value(0 or 1).
        """
        if qubit >= len(self.measure_result) or qubit < 0:
            raise IndexError("qubit index out of range.")
        self.measure_result[qubit] = value

    def get_measure(self):
        """Get the measure result."""
        return self.measure_result

    def get_outcome(self):
        """Get the measure result in binary format."""
        out = self.measure_result[::-1]
        # TODO:improve
        bit_str = "0b"
        for o in out:
            bit_str += str(o)
        return bit_str

    def get_counts(self):
        """Get the number of times the measurement results appear."""
        # TODO:improve
        if self.outcome is None:
            return None

        res = []
        for out in self.outcome:
            res.append({out.bitstr: out.count})
        return json.dumps(res)

    def get_states(self):
        """Get all states"""
        return self.states

    def get_values(self):
        """Get all values"""
        return self.values

    def excute_info(self):
        """The resourece of run."""
        result = {
            "backend": self.backend.name,
            "task_id": self.task_id,
            "status": self.status,
            "arguments": self.arguments,
        }
        return json.dumps(result)



"""Contains back-ends for qutrunk."""
from .backend import Backend
from .local import BackendLocal
from .qusprout import BackendQuSprout, ExecType
from .braket import BackendAWSLocal, BackendAWSDevice
from .ibm import BackendIBM



"""Operator Module."""
from qutrunk.circuit.gates.basicgate import BasicGate


class Operator(BasicGate):
    """A series of quantum gate operations \
        provide a simple and convenient way to construct a quantum algorithm."""

    pass




"""AWS Braket job."""
from typing import Optional, Union

from braket.aws import AwsQuantumTask
from braket.tasks.local_quantum_task import LocalQuantumTask

from qutrunk.backends import Backend


class MeasureResult:
    def __init__(self, id=0, value=0):
        # TODO: id是关键字，不建议使用
        self.id = id
        self.value = value


class OutcomeResult:
    def __init__(self, bit_str="", count=0):
        self.bitstr = bit_str
        self.count = count


class Result:
    def __init__(self):
        self.measureSet = []
        self.outcomeSet = []


class AWSBraketJob:
    """AWSBraketJob."""

    def __init__(
            self,
            job_id: str,
            backend: Backend,
            task: Union[LocalQuantumTask, AwsQuantumTask],
            **metadata: Optional[dict]
    ):
        """AWSBraketJob for local execution of circuits.

        Args:
            job_id: id of the job
            backend: Local simulator
            tasks: Executed tasks
            **metadata:
        """
        self._job_id = job_id
        self._backend = backend
        self._task = task
        self._metadata = metadata

    @property
    def shots(self) -> int:
        """Return the number of shots.

        Returns:
            shots: int with the number of shots.
        """
        return (
            self._metadata["shots"] if "shots" in self._metadata else 0
        )

    def result(self) -> Result:
        # todo bit str big end?
        counter = self._task.result().measurement_counts
        res = Result()
        for bs, cnt in dict(counter).items():
            res.outcomeSet.append(OutcomeResult(bs, cnt))
        return res

    def cancel(self):
        self._task.cancel()

    def status(self):
        return self._task.state()



from qutrunk.circuit import Command
from .basicgate import BasicGate


class MeasureGate(BasicGate):
    """Measures a single qubit, collapsing it randomly to 0 or 1.

    Example:
        .. code-block:: python

            Measure * qr[0]
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "Measure"

    def __or__(self, qubit):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to be measured.

        Example:
            .. code-block:: python

                Measure * qr[0]
        """
        targets = [qubit.index]
        cmd = Command(self, targets)
        self.commit(qubit.circuit, cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubit)


Measure = MeasureGate()



import pytest
from numpy import pi

from qutrunk.circuit import QCircuit
from qutrunk.backends import BackendQuSprout, backend
from qutrunk.circuit.ops import AMP
from qutrunk.test.gate.check_all_state import check_all_state

def test_amplitudes_local():
    """test set amp for local"""
    # 使用本地量子计算模拟器
    qubit_len = 3
    circuit = QCircuit()
    qr = circuit.allocate(qubit_len)

    amplist = [1-2j, 2+3j, 3-4j, 0.5+0.7j]
    startind = 1
    numamps = 2
    AMP(amplist, startind, numamps) * qr

    if numamps > len(amplist):
        numamps = len(amplist)

    res = ['0, 0'] * (2**qubit_len)
    res[0] = '1, 0'

    for i in range(numamps): 
        res[startind] = ','.join([str(qr.circuit.cmds[0].cmdex.amp.reals[i]), str(qr.circuit.cmds[0].cmdex.amp.imags[i])])
        startind += 1

    res_box = circuit.get_statevector()

    # 检查数据是否一致
    assert check_all_state(res, res_box)

test_amplitudes_local()

def test_amplitudes_qusprout():
    """test set amp for local"""
    # 使用本地量子计算模拟器
    qubit_len = 3
    circuit = QCircuit(backend=BackendQuSprout())
    qr = circuit.allocate(qubit_len)

    amplist = [1, 2, 3, 4, 5]
    startind = 3
    numamps = 3
    AMP(amplist, startind, numamps) * qr

    if numamps > len(amplist):
        numamps = len(amplist)

    res = ['0, 0'] * (2**qubit_len)
    res[0] = '1, 0'

    for i in range(numamps): 
        res[startind] = ','.join([str(qr.circuit.cmds[0].cmdex.amp.reals[i]), str(qr.circuit.cmds[0].cmdex.amp.imags[i])])
        startind += 1

    res_box = circuit.get_statevector()

    # 检查数据是否一致
    assert check_all_state(res, res_box)

test_amplitudes_qusprout()


"""Test the load function in QCircuit."""
from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, Measure, CNOT


def test_load_qusl():
    """Test the functionality to deserialize a file object containing an qusl document into a Python object."""
    qc = QCircuit()
    qreg = qc.allocate(2)

    H * qreg[0]
    CNOT * (qreg[0], qreg[1])
    Measure * qreg[0]
    Measure * qreg[1]

    expect_in = qc.cmds

    circuit_in = qc.load(file="a.qusl", format="qusl")

    assert expect_in == circuit_in.cmds


def test_load_openqasm():
    """Test the functionality to deserialize a file object containing an OpenQASM document into a Python object."""
    qc = QCircuit()
    qreg = qc.allocate(2)

    H * qreg[0]
    CNOT * (qreg[0], qreg[1])
    Measure * qreg[0]
    Measure * qreg[1]

    expect_in = qc.cmds

    circuit_in = qc.load(file="b.qasm", format="openqasm")

    assert expect_in == circuit_in.cmds






"""Phase Kickback example."""

from math import pi

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, NOT, Measure, CP


def run_phase_kickback(backend=None):
    # allocate
    qc = QCircuit(backend=backend)
    q1, q2 = qc.allocate([2, 1])

    # apply gate
    H * q1[0]
    H * q1[1]
    NOT * q2[0]
    CP(pi / 4) * (q1[0], q2[0])
    CP(pi / 2) * (q1[1], q2[0])
    Measure * q1[0]
    Measure * q1[1]
    Measure * q2[0]

    # print circuit
    qc.print()

    # run circuit
    res = qc.run(shots=100)

    # print result
    print(res.get_measure())

    return qc


if __name__ == "__main__":
    circuit = run_phase_kickback()
    circuit.draw()




# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""OpenQASM parser."""

import os
import shutil
import tempfile

import numpy as np
from ply import yacc

from . import node
from .exceptions import QasmError
from .qasmlexer import QasmLexer


class QasmParser:
    """OPENQASM Parser."""

    def __init__(self, filename):
        """Create the parser."""
        if filename is None:
            filename = ""
        self.lexer = QasmLexer(filename)
        self.tokens = self.lexer.tokens
        self.parse_dir = tempfile.mkdtemp(prefix="qiskit")
        self.precedence = (
            ("left", "+", "-"),
            ("left", "*", "/"),
            ("left", "negative", "positive"),
            ("right", "^"),
        )
        # For yacc, also, write_tables = Bool and optimize = Bool
        self.parser = yacc.yacc(module=self, debug=False, outputdir=self.parse_dir)
        self.qasm = None
        self.parse_deb = False
        self.global_symtab = {}  # global symtab
        self.current_symtab = self.global_symtab  # top of symbol stack
        self.symbols = []  # symbol stack
        self.external_functions = [
            "sin",
            "cos",
            "tan",
            "exp",
            "ln",
            "sqrt",
            "acos",
            "atan",
            "asin",
        ]

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if os.path.exists(self.parse_dir):
            shutil.rmtree(self.parse_dir)

    def update_symtab(self, obj):
        """Update a node in the symbol table.

        Everything in the symtab must be a node with these attributes:
        name - the string name of the object
        type - the string type of the object
        line - the source line where the type was first found
        file - the source file where the type was first found
        """
        if obj.name in self.current_symtab:
            prev = self.current_symtab[obj.name]
            raise QasmError(
                "Duplicate declaration for",
                obj.type + " '" + obj.name + "' at line",
                str(obj.line) + ", file",
                obj.file + ".\nPrevious occurrence at line",
                str(prev.line) + ", file",
                prev.file,
            )
        self.current_symtab[obj.name] = obj

    def verify_declared_bit(self, obj):
        """Verify a qubit id against the gate prototype."""
        # We are verifying gate args against the formal parameters of a
        # gate prototype.
        if obj.name not in self.current_symtab:
            raise QasmError(
                "Cannot find symbol '" + obj.name + "' in argument list for gate, line",
                str(obj.line),
                "file",
                obj.file,
            )

        # This insures the thing is from the bitlist and not from the
        # argument list.
        sym = self.current_symtab[obj.name]
        if not (sym.type == "id" and sym.is_bit):
            raise QasmError("Bit", obj.name, "is not declared as a bit in the gate.")

    def verify_bit_list(self, obj):
        """Verify each qubit in a list of ids."""
        # We expect the object to be a bitlist or an idlist, we don't care.
        # We will iterate it and ensure everything in it is declared as a bit,
        # and throw if not.
        for children in obj.children:
            self.verify_declared_bit(children)

    def verify_exp_list(self, obj):
        """Verify each expression in a list."""
        # A tad harder.  This is a list of expressions each of which could be
        # the head of a tree. We need to recursively walk each of these and
        # ensure that any Id elements resolve to the current stack.
        #
        # I believe we only have to look at the current symtab.
        if obj.children is not None:
            for children in obj.children:
                if isinstance(children, node.Id):
                    if children.name in self.external_functions:
                        continue

                    if children.name not in self.current_symtab:
                        raise QasmError(
                            "Argument '"
                            + children.name
                            + "' in expression cannot be "
                            + "found, line",
                            str(children.line),
                            "file",
                            children.file,
                        )
                else:
                    if hasattr(children, "children"):
                        self.verify_exp_list(children)

    def verify_as_gate(self, obj, bitlist, arglist=None):
        """Verify a user defined gate call."""
        if obj.name not in self.global_symtab:
            raise QasmError(
                "Cannot find gate definition for '" + obj.name + "', line",
                str(obj.line),
                "file",
                obj.file,
            )
        g_sym = self.global_symtab[obj.name]
        if g_sym.type not in ("gate", "opaque"):
            raise QasmError(
                "'"
                + obj.name
                + "' is used as a gate "
                + "or opaque call but the symbol is neither;"
                + " it is a '"
                + g_sym.type
                + "' line",
                str(obj.line),
                "file",
                obj.file,
            )

        if g_sym.n_bits() != bitlist.size():
            raise QasmError(
                "Gate or opaque call to '" + obj.name + "' uses",
                str(bitlist.size()),
                "qubits but is declared for",
                str(g_sym.n_bits()),
                "qubits",
                "line",
                str(obj.line),
                "file",
                obj.file,
            )

        if arglist:
            if g_sym.n_args() != arglist.size():
                raise QasmError(
                    "Gate or opaque call to '" + obj.name + "' uses",
                    str(arglist.size()),
                    "qubits but is declared for",
                    str(g_sym.n_args()),
                    "qubits",
                    "line",
                    str(obj.line),
                    "file",
                    obj.file,
                )
        else:
            if g_sym.n_args() > 0:
                raise QasmError(
                    "Gate or opaque call to '"
                    + obj.name
                    + "' has no arguments but is declared for",
                    str(g_sym.n_args()),
                    "qubits",
                    "line",
                    str(obj.line),
                    "file",
                    obj.file,
                )

    def verify_reg(self, obj, object_type):
        """Verify a register."""
        # How to verify:
        #    types must match
        #    indexes must be checked
        if obj.name not in self.global_symtab:
            raise QasmError(
                "Cannot find definition for",
                object_type,
                "'" + obj.name + "'",
                "at line",
                str(obj.line),
                "file",
                obj.file,
            )

        g_sym = self.global_symtab[obj.name]

        if g_sym.type != object_type:
            raise QasmError(
                "Type for '"
                + g_sym.name
                + "' should be '"
                + object_type
                + "' but was found to be '"
                + g_sym.type
                + "'",
                "line",
                str(obj.line),
                "file",
                obj.file,
            )

        if obj.type == "indexed_id":
            bound = g_sym.index
            ndx = obj.index
            if ndx < 0 or ndx >= bound:
                raise QasmError(
                    "Register index for '" + g_sym.name + "' out of bounds. Index is",
                    str(ndx),
                    "bound is 0 <= index <",
                    str(bound),
                    "at line",
                    str(obj.line),
                    "file",
                    obj.file,
                )

    def verify_reg_list(self, obj, object_type):
        """Verify a list of registers."""
        # We expect the object to be a bitlist or an idlist, we don't care.
        # We will iterate it and ensure everything in it is declared as a bit,
        # and throw if not.
        for children in obj.children:
            self.verify_reg(children, object_type)

    def id_tuple_list(self, id_node):
        """Return a list of (name, index) tuples for this id node."""
        if id_node.type != "id":
            raise QasmError("internal error, id_tuple_list")
        bit_list = []
        try:
            g_sym = self.current_symtab[id_node.name]
        except KeyError:
            g_sym = self.global_symtab[id_node.name]
        if g_sym.type in ("qreg", "creg"):
            # Return list of (name, idx) for reg ids
            for idx in range(g_sym.index):
                bit_list.append((id_node.name, idx))
        else:
            # Return (name, -1) for other ids
            bit_list.append((id_node.name, -1))
        return bit_list

    def verify_distinct(self, list_of_nodes):
        """Check that objects in list_of_nodes represent distinct (qu)bits.

        list_of_nodes is a list containing nodes of type id, indexed_id,
        primary_list, or id_list. We assume these are all the same type
        'qreg' or 'creg'.
        This method raises an exception if list_of_nodes refers to the
        same object more than once.
        """
        bit_list = []
        line_number = -1
        filename = ""
        for node_ in list_of_nodes:
            # id node: add all bits in register or (name, -1) for id
            if node_.type == "id":
                bit_list.extend(self.id_tuple_list(node_))
                line_number = node_.line
                filename = node_.file
            # indexed_id: add the bit
            elif node_.type == "indexed_id":
                bit_list.append((node_.name, node_.index))
                line_number = node_.line
                filename = node_.file
            # primary_list: for each id or indexed_id child, add
            elif node_.type == "primary_list":
                for child in node_.children:
                    if child.type == "id":
                        bit_list.extend(self.id_tuple_list(child))
                    else:
                        bit_list.append((child.name, child.index))
                    line_number = child.line
                    filename = child.file
            # id_list: for each id, add
            elif node_.type == "id_list":
                for child in node_.children:
                    bit_list.extend(self.id_tuple_list(child))
                    line_number = child.line
                    filename = child.file
            else:
                raise QasmError("internal error, verify_distinct")
        if len(bit_list) != len(set(bit_list)):
            raise QasmError(
                "duplicate identifiers at line %d file %s" % (line_number, filename)
            )

    def pop_scope(self):
        """Return to the previous scope."""
        self.current_symtab = self.symbols.pop()

    def push_scope(self):
        """Enter a new scope."""
        self.symbols.append(self.current_symtab)
        self.current_symtab = {}

    # ---- Begin the PLY parser ----
    start = "main"

    def p_main(self, program):
        """
        main : program
        """
        self.qasm = program[1]

    # ----------------------------------------
    #  program : statement
    #          | program statement
    # ----------------------------------------
    def p_program_0(self, program):
        """
        program : statement
        """
        program[0] = node.Program([program[1]])

    def p_program_1(self, program):
        """
        program : program statement
        """
        program[0] = program[1]
        program[0].add_child(program[2])

    # ----------------------------------------
    #  statement : decl
    #            | quantum_op ';'
    #            | format ';'
    # ----------------------------------------
    def p_statement(self, program):
        """
        statement : decl
                  | quantum_op ';'
                  | format ';'
                  | ignore
                  | quantum_op error
                  | format error
        """
        if len(program) > 2:
            if program[2] != ";":
                raise QasmError(
                    "Missing ';' at end of statement; " + "received",
                    str(program[2].value),
                )
        program[0] = program[1]

    def p_format(self, program):
        """
        format : FORMAT
        """
        version = node.Format(program[1])
        if (version.majorversion != "2") or (version.minorversion != "0"):
            provided_version = f"{version.majorversion}.{version.minorversion}"
            raise QasmError(
                f"Invalid version: '{provided_version}'. This module supports OpenQASM 2.0 only."
            )
        program[0] = version

    # ----------------------------------------
    #  id : ID
    # ----------------------------------------
    def p_id(self, program):
        """
        id : ID
        """
        program[0] = program[1]

    def p_id_e(self, program):
        """
        id : error
        """
        raise QasmError("Expected an ID, received '" + str(program[1].value) + "'")

    # ----------------------------------------
    #  indexed_id : ID [ int ]
    # ----------------------------------------
    def p_indexed_id(self, program):
        """
        indexed_id : id '[' NNINTEGER ']'
                   | id '[' NNINTEGER error
                   | id '[' error
        """
        if len(program) == 4:
            raise QasmError(
                "Expecting an integer index; received", str(program[3].value)
            )
        if program[4] != "]":
            raise QasmError(
                "Missing ']' in indexed ID; received", str(program[4].value)
            )
        program[0] = node.IndexedId([program[1], node.Int(program[3])])

    # ----------------------------------------
    #  primary : id
    #          | indexed_id
    # ----------------------------------------
    def p_primary(self, program):
        """
        primary : id
                | indexed_id
        """
        program[0] = program[1]

    # ----------------------------------------
    #  id_list : id
    #          | id_list ',' id
    # ----------------------------------------
    def p_id_list_0(self, program):
        """
        id_list : id
        """
        program[0] = node.IdList([program[1]])

    def p_id_list_1(self, program):
        """
        id_list : id_list ',' id
        """
        program[0] = program[1]
        program[0].add_child(program[3])

    # ----------------------------------------
    #  gate_id_list : id
    #               | gate_id_list ',' id
    # ----------------------------------------
    def p_gate_id_list_0(self, program):
        """
        gate_id_list : id
        """
        program[0] = node.IdList([program[1]])
        self.update_symtab(program[1])

    def p_gate_id_list_1(self, program):
        """
        gate_id_list : gate_id_list ',' id
        """
        program[0] = program[1]
        program[0].add_child(program[3])
        self.update_symtab(program[3])

    # ----------------------------------------
    #  bit_list : bit
    #           | bit_list ',' bit
    # ----------------------------------------
    def p_bit_list_0(self, program):
        """
        bit_list : id
        """
        program[0] = node.IdList([program[1]])
        program[1].is_bit = True
        self.update_symtab(program[1])

    def p_bit_list_1(self, program):
        """
        bit_list : bit_list ',' id
        """
        program[0] = program[1]
        program[0].add_child(program[3])
        program[3].is_bit = True
        self.update_symtab(program[3])

    # ----------------------------------------
    #  primary_list : primary
    #               | primary_list ',' primary
    # ----------------------------------------
    def p_primary_list_0(self, program):
        """
        primary_list : primary
        """
        program[0] = node.PrimaryList([program[1]])

    def p_primary_list_1(self, program):
        """
        primary_list : primary_list ',' primary
        """
        program[0] = program[1]
        program[1].add_child(program[3])

    # ----------------------------------------
    #  decl : qreg_decl
    #       | creg_decl
    #       | gate_decl
    # ----------------------------------------
    def p_decl(self, program):
        """
        decl : qreg_decl ';'
             | creg_decl ';'
             | qreg_decl error
             | creg_decl error
             | gate_decl
        """
        if len(program) > 2:
            if program[2] != ";":
                raise QasmError(
                    "Missing ';' in qreg or creg declaration."
                    " Instead received '" + program[2].value + "'"
                )
        program[0] = program[1]

    # ----------------------------------------
    #  qreg_decl : QREG indexed_id
    # ----------------------------------------
    def p_qreg_decl(self, program):
        """
        qreg_decl : QREG indexed_id
        """
        program[0] = node.Qreg([program[2]])
        if program[2].name in self.external_functions:
            raise QasmError(
                "QREG names cannot be reserved words. "
                + "Received '"
                + program[2].name
                + "'"
            )
        if program[2].index == 0:
            raise QasmError("QREG size must be positive")
        self.update_symtab(program[0])

    def p_qreg_decl_e(self, program):
        """
        qreg_decl : QREG error
        """
        raise QasmError(
            "Expecting indexed id (ID[int]) in QREG" + " declaration; received",
            program[2].value,
        )

    # ----------------------------------------
    #  creg_decl : QREG indexed_id
    # ----------------------------------------
    def p_creg_decl(self, program):
        """
        creg_decl : CREG indexed_id
        """
        program[0] = node.Creg([program[2]])
        if program[2].name in self.external_functions:
            raise QasmError(
                "CREG names cannot be reserved words. "
                + "Received '"
                + program[2].name
                + "'"
            )
        if program[2].index == 0:
            raise QasmError("CREG size must be positive")
        self.update_symtab(program[0])

    def p_creg_decl_e(self, program):
        """
        creg_decl : CREG error
        """
        raise QasmError(
            "Expecting indexed id (ID[int]) in CREG" + " declaration; received",
            program[2].value,
        )

    # Gate_body will throw if there are errors, so we don't need to cover
    # that here. Same with the id_lists - if they are not legal, we die
    # before we get here
    #
    # ----------------------------------------
    #  gate_decl : GATE id gate_scope                      bit_list gate_body
    #            | GATE id gate_scope '(' ')'              bit_list gate_body
    #            | GATE id gate_scope '(' gate_id_list ')' bit_list gate_body
    #
    # ----------------------------------------
    def p_gate_decl_0(self, program):
        """
        gate_decl : GATE id gate_scope bit_list gate_body
        """
        program[0] = node.Gate([program[2], program[4], program[5]])
        if program[2].name in self.external_functions:
            raise QasmError(
                "GATE names cannot be reserved words. "
                + "Received '"
                + program[2].name
                + "'"
            )
        self.pop_scope()
        self.update_symtab(program[0])

    def p_gate_decl_1(self, program):
        """
        gate_decl : GATE id gate_scope '(' ')' bit_list gate_body
        """
        program[0] = node.Gate([program[2], program[6], program[7]])
        if program[2].name in self.external_functions:
            raise QasmError(
                "GATE names cannot be reserved words. "
                + "Received '"
                + program[2].name
                + "'"
            )
        self.pop_scope()
        self.update_symtab(program[0])

    def p_gate_decl_2(self, program):
        """
        gate_decl : GATE id gate_scope '(' gate_id_list ')' bit_list gate_body
        """
        program[0] = node.Gate([program[2], program[5], program[7], program[8]])
        if program[2].name in self.external_functions:
            raise QasmError(
                "GATE names cannot be reserved words. "
                + "Received '"
                + program[2].name
                + "'"
            )
        self.pop_scope()
        self.update_symtab(program[0])

    def p_gate_scope(self, _):
        """
        gate_scope :
        """
        self.push_scope()

    # ----------------------------------------
    #  gate_body : '{' gate_op_list '}'
    #            | '{' '}'
    #
    #            | '{' gate_op_list error
    #            | '{' error
    #
    # Error handling: gete_op will throw if there's a problem so we won't
    #                 get here with in the gate_op_list
    # ----------------------------------------
    def p_gate_body_0(self, program):
        """
        gate_body : '{' '}'
        """
        if program[2] != "}":
            raise QasmError(
                "Missing '}' in gate definition; received'"
                + str(program[2].value)
                + "'"
            )
        program[0] = node.GateBody(None)

    def p_gate_body_1(self, program):
        """
        gate_body : '{' gate_op_list '}'
        """
        program[0] = node.GateBody(program[2])

    # ----------------------------------------
    #  gate_op_list : gate_op
    #               | gate_op_ist gate_op
    #
    # Error handling: gete_op will throw if there's a problem so we won't
    #                 get here with errors
    # ----------------------------------------
    def p_gate_op_list_0(self, program):
        """
        gate_op_list : gate_op
        """
        program[0] = [program[1]]

    def p_gate_op_list_1(self, program):
        """
        gate_op_list : gate_op_list gate_op
        """
        program[0] = program[1]
        program[0].append(program[2])

    # ----------------------------------------
    # These are for use outside of gate_bodies and allow
    # indexed ids everywhere.
    #
    # unitary_op : U '(' exp_list ')'  primary
    #            | CX                  primary ',' primary
    #            | id                  primary_list
    #            | id '(' ')'          primary_list
    #            | id '(' exp_list ')' primary_list
    #
    # Note that it might not be unitary - this is the mechanism that
    # is also used to invoke calls to 'opaque'
    # ----------------------------------------
    def p_unitary_op_0(self, program):
        """
        unitary_op : U '(' exp_list ')' primary
        """
        program[0] = node.UniversalUnitary([program[3], program[5]])
        self.verify_reg(program[5], "qreg")
        self.verify_exp_list(program[3])

    def p_unitary_op_1(self, program):
        """
        unitary_op : CX primary ',' primary
        """
        program[0] = node.Cnot([program[2], program[4]])
        self.verify_reg(program[2], "qreg")
        self.verify_reg(program[4], "qreg")
        self.verify_distinct([program[2], program[4]])
        # TODO: check that if both primary are id, same size
        # TODO: this needs to be checked in other cases too

    def p_unitary_op_2(self, program):
        """
        unitary_op : id primary_list
        """
        program[0] = node.CustomUnitary([program[1], program[2]])
        self.verify_as_gate(program[1], program[2])
        self.verify_reg_list(program[2], "qreg")
        self.verify_distinct([program[2]])

    def p_unitary_op_3(self, program):
        """
        unitary_op : id '(' ')' primary_list
        """
        program[0] = node.CustomUnitary([program[1], program[4]])
        self.verify_as_gate(program[1], program[4])
        self.verify_reg_list(program[4], "qreg")
        self.verify_distinct([program[4]])

    def p_unitary_op_4(self, program):
        """
        unitary_op : id '(' exp_list ')' primary_list
        """
        program[0] = node.CustomUnitary([program[1], program[3], program[5]])
        self.verify_as_gate(program[1], program[5], arglist=program[3])
        self.verify_reg_list(program[5], "qreg")
        self.verify_exp_list(program[3])
        self.verify_distinct([program[5]])

    # ----------------------------------------
    # This is a restricted set of "quantum_op" which also
    # prohibits indexed ids, for use in a gate_body
    #
    # gate_op : U '(' exp_list ')'  id         ';'
    #         | CX                  id ',' id  ';'
    #         | id                  id_list    ';'
    #         | id '(' ')'          id_list    ';'
    #         | id '(' exp_list ')' id_list    ';'
    #         | BARRIER id_list                ';'
    # ----------------------------------------
    def p_gate_op_0(self, program):
        """
        gate_op : U '(' exp_list ')' id ';'
        """
        program[0] = node.UniversalUnitary([program[3], program[5]])
        self.verify_declared_bit(program[5])
        self.verify_exp_list(program[3])

    def p_gate_op_0e1(self, p):
        """
        gate_op : U '(' exp_list ')' error
        """
        raise QasmError("Invalid U inside gate definition. " + "Missing bit id or ';'")

    def p_gate_op_0e2(self, _):
        """
        gate_op : U '(' exp_list error
        """
        raise QasmError("Missing ')' in U invocation in gate definition.")

    def p_gate_op_1(self, program):
        """
        gate_op : CX id ',' id ';'
        """
        program[0] = node.Cnot([program[2], program[4]])
        self.verify_declared_bit(program[2])
        self.verify_declared_bit(program[4])
        self.verify_distinct([program[2], program[4]])

    def p_gate_op_1e1(self, program):
        """
        gate_op : CX error
        """
        raise QasmError(
            "Invalid CX inside gate definition. "
            + "Expected an ID or ',', received '"
            + str(program[2].value)
            + "'"
        )

    def p_gate_op_1e2(self, program):
        """
        gate_op : CX id ',' error
        """
        raise QasmError(
            "Invalid CX inside gate definition. "
            + "Expected an ID or ';', received '"
            + str(program[4].value)
            + "'"
        )

    def p_gate_op_2(self, program):
        """
        gate_op : id id_list ';'
        """
        program[0] = node.CustomUnitary([program[1], program[2]])
        # To verify:
        # 1. id is declared as a gate in global scope
        # 2. everything in the id_list is declared as a bit in local scope
        self.verify_as_gate(program[1], program[2])
        self.verify_bit_list(program[2])
        self.verify_distinct([program[2]])

    def p_gate_op_2e(self, _):
        """
        gate_op : id  id_list error
        """
        raise QasmError("Invalid gate invocation inside gate definition.")

    def p_gate_op_3(self, program):
        """
        gate_op : id '(' ')' id_list ';'
        """
        program[0] = node.CustomUnitary([program[1], program[4]])
        self.verify_as_gate(program[1], program[4])
        self.verify_bit_list(program[4])
        self.verify_distinct([program[4]])

    def p_gate_op_4(self, program):
        """
        gate_op : id '(' exp_list ')' id_list ';'
        """
        program[0] = node.CustomUnitary([program[1], program[3], program[5]])
        self.verify_as_gate(program[1], program[5], arglist=program[3])
        self.verify_bit_list(program[5])
        self.verify_exp_list(program[3])
        self.verify_distinct([program[5]])

    def p_gate_op_4e0(self, _):
        """
        gate_op : id '(' ')'  error
        """
        raise QasmError("Invalid bit list inside gate definition or" + " missing ';'")

    def p_gate_op_4e1(self, _):
        """
        gate_op : id '('   error
        """
        raise QasmError("Unmatched () for gate invocation inside gate" + " invocation.")

    def p_gate_op_5(self, program):
        """
        gate_op : BARRIER id_list ';'
        """
        program[0] = node.Barrier([program[2]])
        self.verify_bit_list(program[2])
        self.verify_distinct([program[2]])

    def p_gate_op_5e(self, _):
        """
        gate_op : BARRIER error
        """
        raise QasmError("Invalid barrier inside gate definition.")

    # ----------------------------------------
    # opaque : OPAQUE id gate_scope                      bit_list
    #        | OPAQUE id gate_scope '(' ')'              bit_list
    #        | OPAQUE id gate_scope '(' gate_id_list ')' bit_list
    #
    # These are like gate declarations only without a body.
    # ----------------------------------------
    def p_opaque_0(self, program):
        """
        opaque : OPAQUE id gate_scope bit_list
        """
        # TODO: Review Opaque function
        program[0] = node.Opaque([program[2], program[4]])
        if program[2].name in self.external_functions:
            raise QasmError(
                "OPAQUE names cannot be reserved words. "
                + "Received '"
                + program[2].name
                + "'"
            )
        self.pop_scope()
        self.update_symtab(program[0])

    def p_opaque_1(self, program):
        """
        opaque : OPAQUE id gate_scope '(' ')' bit_list
        """
        program[0] = node.Opaque([program[2], program[6]])
        self.pop_scope()
        self.update_symtab(program[0])

    def p_opaque_2(self, program):
        """
        opaque : OPAQUE id gate_scope '(' gate_id_list ')' bit_list
        """
        program[0] = node.Opaque([program[2], program[5], program[7]])
        if program[2].name in self.external_functions:
            raise QasmError(
                "OPAQUE names cannot be reserved words. "
                + "Received '"
                + program[2].name
                + "'"
            )
        self.pop_scope()
        self.update_symtab(program[0])

    def p_opaque_1e(self, _):
        """
        opaque : OPAQUE id gate_scope '(' error
        """
        raise QasmError("Poorly formed OPAQUE statement.")

    # ----------------------------------------
    # measure : MEASURE primary ASSIGN primary
    # ----------------------------------------
    def p_measure(self, program):
        """
        measure : MEASURE primary ASSIGN primary
        """
        program[0] = node.Measure([program[2], program[4]])
        self.verify_reg(program[2], "qreg")
        self.verify_reg(program[4], "creg")

    def p_measure_e(self, program):
        """
        measure : MEASURE primary error
        """
        raise QasmError("Illegal measure statement." + str(program[3].value))

    # ----------------------------------------
    # barrier : BARRIER primary_list
    #
    # Errors are covered by handling errors in primary_list
    # ----------------------------------------
    def p_barrier(self, program):
        """
        barrier : BARRIER primary_list
        """
        program[0] = node.Barrier([program[2]])
        self.verify_reg_list(program[2], "qreg")
        self.verify_distinct([program[2]])

    # ----------------------------------------
    # reset : RESET primary
    # ----------------------------------------
    def p_reset(self, program):
        """
        reset : RESET primary
        """
        program[0] = node.Reset([program[2]])
        self.verify_reg(program[2], "qreg")

    # ----------------------------------------
    # IF '(' ID MATCHES NNINTEGER ')' quantum_op
    # ----------------------------------------
    def p_if(self, program):
        """
        if : IF '(' id MATCHES NNINTEGER ')' quantum_op
        if : IF '(' id error
        if : IF '(' id MATCHES error
        if : IF '(' id MATCHES NNINTEGER error
        if : IF error
        """
        if len(program) == 3:
            raise QasmError("Ill-formed IF statement. Perhaps a" + " missing '('?")
        if len(program) == 5:
            raise QasmError(
                "Ill-formed IF statement.  Expected '==', "
                + "received '"
                + str(program[4].value)
            )
        if len(program) == 6:
            raise QasmError(
                "Ill-formed IF statement.  Expected a number, "
                + "received '"
                + str(program[5].value)
            )
        if len(program) == 7:
            raise QasmError("Ill-formed IF statement, unmatched '('")

        if program[7].type == "if":
            raise QasmError("Nested IF statements not allowed")
        if program[7].type == "barrier":
            raise QasmError("barrier not permitted in IF statement")

        program[0] = node.If([program[3], node.Int(program[5]), program[7]])

    # ----------------------------------------
    # These are all the things you can have outside of a gate declaration
    #        quantum_op : unitary_op
    #                   | opaque
    #                   | measure
    #                   | reset
    #                   | barrier
    #                   | if
    #
    # ----------------------------------------
    def p_quantum_op(self, program):
        """
        quantum_op : unitary_op
                   | opaque
                   | measure
                   | barrier
                   | reset
                   | if
        """
        program[0] = program[1]

    # ----------------------------------------
    # unary : NNINTEGER
    #       | REAL
    #       | PI
    #       | ID
    #       | '(' expression ')'
    #       | id '(' expression ')'
    #
    # We will trust 'expression' to throw before we have to handle it here
    # ----------------------------------------
    def p_unary_0(self, program):
        """
        unary : NNINTEGER
        """
        program[0] = node.Int(program[1])

    def p_unary_1(self, program):
        """
        unary : REAL
        """
        program[0] = node.Real(program[1])

    def p_unary_2(self, program):
        """
        unary : PI
        """
        program[0] = node.Real(np.pi)

    def p_unary_3(self, program):
        """
        unary : id
        """
        program[0] = program[1]

    def p_unary_4(self, program):
        """
        unary : '(' expression ')'
        """
        program[0] = program[2]

    def p_unary_6(self, program):
        """
        unary : id '(' expression ')'
        """
        # note this is a semantic check, not syntactic
        if program[1].name not in self.external_functions:
            raise QasmError("Illegal external function call: ", str(program[1].name))
        program[0] = node.External([program[1], program[3]])

    # ----------------------------------------
    # Prefix
    # ----------------------------------------

    def p_expression_1(self, program):
        """
        expression : '-' expression %prec negative
                    | '+' expression %prec positive
        """
        program[0] = node.Prefix([node.UnaryOperator(program[1]), program[2]])

    def p_expression_0(self, program):
        """
        expression : expression '*' expression
                    | expression '/' expression
                    | expression '+' expression
                    | expression '-' expression
                    | expression '^' expression
        """
        program[0] = node.BinaryOp(
            [node.BinaryOperator(program[2]), program[1], program[3]]
        )

    def p_expression_2(self, program):
        """
        expression : unary
        """
        program[0] = program[1]

    # ----------------------------------------
    # exp_list : exp
    #          | exp_list ',' exp
    # ----------------------------------------
    def p_exp_list_0(self, program):
        """
        exp_list : expression
        """
        program[0] = node.ExpressionList([program[1]])

    def p_exp_list_1(self, program):
        """
        exp_list : exp_list ',' expression
        """
        program[0] = program[1]
        program[0].add_child(program[3])

    def p_ignore(self, _):
        """
        ignore : STRING
        """
        # this should never hit but it keeps the insuppressible warnings at bay
        pass

    def p_error(self, program):
        # EOF is a special case because the stupid error token isn't placed
        # on the stack
        if not program:
            raise QasmError("Error at end of file. " + "Perhaps there is a missing ';'")

        col = self.find_column(self.lexer.data, program)
        print("Error near line", str(self.lexer.lineno), "Column", col)

    def find_column(self, input_, token):
        """Compute the column.

        Input is the input text string.
        token is a token instance.
        """
        if token is None:
            return 0
        last_cr = input_.rfind("\n", 0, token.lexpos)
        last_cr = max(last_cr, 0)
        column = (token.lexpos - last_cr) + 1
        return column

    def read_tokens(self):
        """finds and reads the tokens."""
        try:
            while True:
                token = self.lexer.token()

                if not token:
                    break

                yield token
        except QasmError as e:
            print("Exception tokenizing qasm file:", e.msg)

    def parse_debug(self, val):
        """Set the parse_deb field."""
        if val is True:
            self.parse_deb = True
        elif val is False:
            self.parse_deb = False
        else:
            raise QasmError(
                "Illegal debug value '" + str(val) + "' must be True or False."
            )

    def parse(self, data):
        """Parse some data."""
        self.parser.parse(data, lexer=self.lexer, debug=self.parse_deb)
        if self.qasm is None:
            raise QasmError(
                "Uncaught exception in parser; " + "see previous messages for details."
            )
        return self.qasm

    def print_tree(self):
        """Print parsed OPENQASM."""
        if self.qasm is not None:
            self.qasm.to_string(0)
        else:
            print("No parsed qasm to print")

    def run(self, data):
        """Parser runner.

        To use this module stand-alone.
        """
        ast = self.parser.parse(data, debug=True)
        self.parser.parse(data, debug=True)
        ast.to_string(0)



"""Bell state example."""

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, CNOT, Measure


def run_bell_pair(backend=None):
    # Create quantum circuit
    qc = QCircuit(backend=backend)

    # Allocate quantum qubits
    qr = qc.allocate(2)

    # Apply quantum gates
    H * qr[0]
    CNOT * (qr[0], qr[1])
    Measure * qr[0]
    Measure * qr[1]

    # Print quantum circuit
    qc.print()
    # qc.dump(file="bell_pair.qasm", format="openqasm")

    # Run quantum circuit with 100 times
    res = qc.run(shots=100)

    # Print result like:
    #[1, 1]
    #[{"00": 50}, {"11": 50}]
    print(res.get_measure())
    print(res.get_counts())

    # Print quantum circuit exection information
    print(res.excute_info())

    return qc


if __name__ == "__main__":
    # Run locally
    circuit = run_bell_pair()

    # Draw quantum circuit
    circuit.draw()



"""Grover's search algorithm."""

import math
import random

from numpy import pi

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import Measure, All
from qutrunk.circuit.ops import PLUS, QAA


def run_grover(qubits=10, backend=None):
    # Quantum qubits
    num_qubits = qubits

    # Number of amplitudes
    num_elems = 2**num_qubits

    # Count of iteration
    num_reps = math.ceil(pi / 4 * math.sqrt(num_elems))
    print("num_qubits:", num_qubits, "num_elems:", num_elems, "num_reps:", num_reps)

    # Choose target state randomly
    sol_elem = random.randint(0, num_elems - 1)
    print(f"target state: |{str(sol_elem)}>")

    # Create quantum circuit with local python simulator
    circuit = QCircuit(backend=backend, resource=True)

    # Allocate quantum qubits
    qureg = circuit.allocate(num_qubits)

    # Set inital amplitudes to plus state
    PLUS * qureg

    # Apply quantum operator(gates)
    QAA(num_reps, sol_elem) * qureg

    # Measure for all qubits
    All(Measure) * qureg

    # Run circuit in local simulator
    res = circuit.run()

    # Get measure result and print as int
    out = res.get_outcome()
    print("measure result: " + str(int(out, base=2)))

    # Print quantum circuit resource information
    circuit.show_resource()

    # Print quantum circuit execution information
    print(res.excute_info())

    return circuit


if __name__ == "__main__":
    # Run locally
    circuit = run_grover()



"""IBM Backend"""

import math

from qutrunk.backends import Backend
from qutrunk.circuit.gates import CNOT, H, Measure, Rx, Ry, Rz
from .ibm_client import send


# TODO:The IBM quantum chip can only do U1,U2,U3,barriers, and CX / CNOT.
class BackendIBM(Backend):
    """IBM Backend.

    The BackendIBM class, which stores the circuit, transforms it to JSON,
    and sends the circuit through the IBM API.

    Args:
        token: Token used for authorization when send quantum circuit to IBM Backend.
        device: Device type, default: ibmq_qasm_simulator.

    Example:
        .. code-block:: python

            from qutrunk.circuit import QCircuit
            from qutrunk.backends import BackendIBM
            from qutrunk.circuit.gates import H, CNOT, Measure

            # use BackendIBM
            token = "IBM Quantum token"
            qc = QCircuit(backend=BackendIBM(token=token))
            qr = qc.allocate(2)

            # apply gate
            H * qr[0]
            CNOT * (qr[0], qr[1])
            Measure * qr[0]
            Measure * qr[1]

            # run circuit
            res = qc.run(shots=100)
            print(res)
    """

    def __init__(self, token, device=None):
        super().__init__()
        self.circuit = None
        self._token = token
        # default device: "ibmq_qasm_simulator"
        self.device = "ibmq_qasm_simulator"
        # the qubits to allocated
        self._allocated_qubits = set()
        # quantum circuit for json format
        self._json = []
        # measured qubit id
        self._measured_ids = []

    def send_circuit(self, circuit, final=False):
        """Send the quantum circuit to IBM backend.

        Args:
            circuit: Quantum circuit to send.
            final: True if quantum circuit finish, default False.

        Returns:
            The result return from IBM Backend.
        """
        self._allocated_qubits.add(len(circuit.qreg))

        for ct in circuit.cmds:
            self._circuit_to_json(ct)

        for measured_id in self._measured_ids:
            self._json.append(
                {"qubits": [measured_id], "name": "measure", "memory": [measured_id]}
            )
        # print("send circuit:", self._json)

    def _circuit_to_json(self, cmd):
        """Translates the command and in a local variable.

        Args:
            cmd: The command convert to json format.
        """
        gate = cmd.gate

        if gate is Measure:
            self._measured_ids += cmd.targets
        elif gate is CNOT and len(cmd.controls) == 1:
            self._json.append({"qubits": [*cmd.controls, *cmd.targets], "name": "cx"})
        elif gate is H:
            self._json.append(
                {"qubits": cmd.targets, "name": "u2", "params": [0, 3.141592653589793]}
            )
        elif isinstance(gate, (Rx, Ry, Rz)):
            u_name = {"Rx": "u3", "Ry": "u3", "Rz": "u1"}
            u_angle = {
                "Rx": [gate.angle, -math.pi / 2, math.pi / 2],
                "Ry": [gate.angle, 0, 0],
                "Rz": [gate.angle],
            }

            gate_name = u_name[str(gate)[0:2]]
            params = u_angle[str(gate)[0:2]]

            self._json.append(
                {"qubits": cmd.targets, "name": gate_name, "params": params}
            )

        else:
            raise Exception(
                "This command is not currently supported.\n"
                + "The IBM quantum chip can only do U1,U2,U3,barriers, and CX / CNOT."
            )

    def run(self, shots=1):
        """Run quantum circuit.

        Args:
            shots: Circuit run times, for sampling, default: 1.

        Returns:
            JSON: The Result object contain circuit running outcome.
        """
        info = {}
        info["json"] = self._json

        info["nq"] = sum(self._allocated_qubits)
        info["shots"] = shots
        info["maxCredits"] = 10
        info["backend"] = {"name": self.device}

        result = send(
            info,
            device=self.device,
            token=self._token,
            num_retries=shots,
            verbose=True,
        )
        return result

    @property
    def name(self):
        """The name of Backend."""
        return "BackendIBM"



"""Classical register reference object."""

import itertools

from .classical_bit import CBit


class CReg:
    """A classical register.

    Args:
        name (str): The name of the register.
        size: The size of the register.
        circuit: A quantum circuit.
    """

    prefix = "c"
    bit_type = "CBit"
    instances_counter = itertools.count()

    def __init__(self, circuit=None, name=None, size=None):
        if name is None:
            name = f"{self.prefix}{self.instances_counter}"
        self.name = name

        self.cbits = []
        # TODO: delete size ?
        self.size = size
        self.circuit = circuit

        for index in range(self.size):
            self.cbits.append(CBit(self, index))

    def append(self, cbit):
        """Add a CBit instance.

        Args:
            cbit: A list which store CBit instance.
        """
        self.cbits.append(cbit)

    def qasm(self):
        """Return OPENQASM string for this register."""
        return f"creg {self.name}[{self.size}];"

    def __getitem__(self, idx):
        """Return a CBit instance.

        Args:
            idx: The index of CBit.

        Returns:
            CBit instance.
        """
        if not isinstance(idx, int):
            raise ValueError("expected integer index into register")
        return self.cbits[idx]

    def __len__(self):
        return len(self.cbits)

    def index(self, cbit):
        """Find the index of the provided cbit within this register.

        Args:
            cbit: A list which store CBit instance.
        """
        return self.cbits.index(cbit)

    def __setitem__(self, key, value):
        self.cbits[key] = value

    def __repr__(self):
        return f"{self.__class__.__qualname__}({self.size})"



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""OPENQASM Lexer.

This is a wrapper around the PLY lexer to support the "include" statement
by creating a stack of lexers.
"""

import os

import numpy as np
from ply import lex

from . import node
from .exceptions import QasmError

CORE_LIBS_PATH = os.path.join(os.path.dirname(__file__), "libs")
CORE_LIBS = os.listdir(CORE_LIBS_PATH)


class QasmLexer:
    """OPENQASM Lexer.

    This is a wrapper around the PLY lexer to support the "include" statement
    by creating a stack of lexers.
    """

    def __mklexer__(self, filename):
        """Create a PLY lexer."""
        self.lexer = lex.lex(module=self, debug=False)
        self.filename = filename
        self.lineno = 1

        if filename:
            with open(filename) as ifile:
                self.data = ifile.read()
            self.lexer.input(self.data)

    def __init__(self, filename):
        """Create the OPENQASM lexer."""
        self.__mklexer__(filename)
        self.stack = []

    def input(self, data):
        """Set the input text data."""
        self.data = data
        self.lexer.input(data)

    def token(self):
        """Return the next token."""
        ret = self.lexer.token()
        return ret

    def pop(self):
        """Pop a PLY lexer off the stack."""
        self.lexer = self.stack.pop()
        self.filename = self.lexer.qasm_file
        self.lineno = self.lexer.qasm_line

    def push(self, filename):
        """Push a PLY lexer on the stack to parse filename."""
        self.lexer.qasm_file = self.filename
        self.lexer.qasm_line = self.lineno
        self.stack.append(self.lexer)
        self.__mklexer__(filename)

    # ---- Beginning of the PLY lexer ----
    literals = r'=()[]{};<>,.+-/*^"'
    reserved = {
        "barrier": "BARRIER",
        "creg": "CREG",
        "gate": "GATE",
        "if": "IF",
        "measure": "MEASURE",
        "opaque": "OPAQUE",
        "qreg": "QREG",
        "pi": "PI",
        "reset": "RESET",
    }
    tokens = [
        "NNINTEGER",
        "REAL",
        "CX",
        "U",
        "FORMAT",
        "ASSIGN",
        "MATCHES",
        "ID",
        "STRING",
    ] + list(reserved.values())

    def t_REAL(self, t):
        r"(([0-9]+|([0-9]+)?\.[0-9]+|[0-9]+\.)[eE][+-]?[0-9]+)|(([0-9]+)?\.[0-9]+|[0-9]+\.)"
        if np.iscomplex(t):
            return t.real
        else:
            return t

    def t_NNINTEGER(self, t):
        r"[1-9]+[0-9]*|0"
        t.value = int(t.value)
        return t

    def t_ASSIGN(self, t):
        "->"
        return t

    def t_MATCHES(self, t):
        "=="
        return t

    def t_STRING(self, t):
        r"\"([^\\\"]|\\.)*\" "
        return t

    def t_INCLUDE(self, _):
        "include"
        # Now eat up the next two tokens which must be
        # 1 - the name of the include file, and
        # 2 - a terminating semicolon
        #
        # Then push the current lexer onto the stack, create a new one from
        # the include file, and push it onto the stack.
        #
        # When we hit eof (the t_eof) rule, we pop.
        next_token = self.lexer.token()
        lineno = next_token.lineno
        if isinstance(next_token.value, str):
            incfile = next_token.value.strip('"')
        else:
            raise QasmError("Invalid include: must be a quoted string.")

        if incfile in CORE_LIBS:
            incfile = os.path.join(CORE_LIBS_PATH, incfile)

        next_token = self.lexer.token()
        if next_token is None or next_token.value != ";":
            raise QasmError('Invalid syntax, missing ";" at line', str(lineno))

        if not os.path.exists(incfile):
            raise QasmError(
                "Include file %s cannot be found, line %s, file %s"
                % (incfile, str(next_token.lineno), self.filename)
            )
        self.push(incfile)
        return self.lexer.token()

    def t_FORMAT(self, t):
        r"OPENQASM\s+[0-9]+(\.[0-9]+)?"
        return t

    def t_COMMENT(self, _):
        r"//.*"
        pass

    def t_CX(self, t):
        "CX"
        return t

    def t_U(self, t):
        "U"
        return t

    def t_ID(self, t):
        r"[a-z][a-zA-Z0-9_]*"

        t.type = self.reserved.get(t.value, "ID")
        if t.type == "ID":
            t.value = node.Id(t.value, self.lineno, self.filename)
        return t

    def t_newline(self, t):
        r"\n+"
        self.lineno += len(t.value)
        t.lexer.lineno = self.lineno

    def t_eof(self, _):
        if self.stack:
            self.pop()
            return self.lexer.token()
        return None

    t_ignore = " \t\r"

    def t_error(self, t):
        raise QasmError(
            "Unable to match any token rule, got -->%s<-- "
            "Check your OPENQASM source and any include statements." % t.value[0]
        )



import time
from functools import wraps


def timefn(fn):
    """The execution time of function."""

    @wraps(fn)
    def measure_time(*args, **kwargs):
        start_time = time.time()
        result = fn(*args, **kwargs)
        end_time = time.time()
        total_time = end_time - start_time
        return result, total_time

    return measure_time



"""addition example."""

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import Measure, All
from qutrunk.circuit.ops import Classical
from qutrunk.circuit.ops import ADD


def run_addition(num_qubits, init_value, number=0):
    circuit = QCircuit()
    qr = circuit.allocate(num_qubits)

    Classical(init_value) * qr

    ADD(number) * qr

    All(Measure) * qr
    res = circuit.run()
    print(res.get_outcome())

    return circuit


if __name__ == "__main__":
    circuit = run_addition(4, 3, 3)




import numpy as np

from .basicgate import BasicGate
from qutrunk.circuit import Command
from qutrunk.circuit.qubit import QuBit


class X1Gate(BasicGate):
    """Apply the single-qubit X1 gate.

    Example:
        .. code-block:: python

            X1 * qr[0]
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "X1"

    def __or__(self, qubit):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to aplly X1 gate.

        Example:
            .. code-block:: python

                X1 * qr[0]

        Raises:
            NotImplementedError: If the argument is not a Qubit object.
        """
        if not isinstance(qubit, QuBit):
            raise TypeError("The argument must be Qubit object.")

        targets = [qubit.index]
        cmd = Command(self, targets, inverse=self.is_inverse)
        self.commit(qubit.circuit, cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement."""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        factor = 1 / np.sqrt(2)
        return np.matrix([[factor, -1j * factor], [-1j * factor, factor]])

    def inv(self):
        """Apply inverse gate."""
        gate = X1Gate()
        gate.is_inverse = not self.is_inverse
        return gate


X1 = X1Gate()



class QuTrunkError(Exception):
    """Base class for errors raised by QuTrunk."""

    def __init__(self, *message):
        """Set the error message.

        Args:
            message: Error message.
        """
        super().__init__(" ".join(message))
        self.message = " ".join(message)

    def __str__(self):
        """Return the message."""
        return repr(self.message)



"""Bernstein-Vazirani example."""
import random

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import CNOT, X, H
from qutrunk.circuit.ops import Classical

PRECISION = 1e-10


def _apply_oracle(qr, num_qubits, secret):
    bits = secret
    for q in range(1, num_qubits):
        # extract the (q-1)-th bit of secret
        bit = int(bits % 2)
        bits = int(bits / 2)

        # NOT the ancilla, controlling on the q-th qubit
        if bit:
            CNOT * (qr[q], qr[0])


def apply_bernstein_vazirani(num_qubits, secret):
    # create circuit and qureg
    circuit = QCircuit()
    qr = circuit.allocate(num_qubits)

    # start in |0>
    Classical(0) * qr

    # NOT the ancilla
    X * qr[0]

    # H all qubits, including the ancilla
    for q in range(num_qubits):
        H * qr[q]

    _apply_oracle(qr, num_qubits, secret)

    # H all qubits, including the ancilla
    for q in range(num_qubits):
        H * qr[q]

    # infer the output basis state
    ind = 2 * secret + 1
    prob = circuit.get_prob(ind)
    if 1 - prob < PRECISION:
        prob = 1

    print(f"success probability:  {prob}")

    return circuit


if __name__ == "__main__":
    # number of qubits
    num_qubits = 15

    # number of all states
    num_elems = 2 ** (num_qubits - 1)

    # randomly choose the secret parameter
    secret = random.randint(0, num_elems - 1)

    # search for s using BV's algorithm
    circuit = apply_bernstein_vazirani(num_qubits, secret)




"""
DAG Node.
"""


class DAGNode:
    """Parent class for DAGOpNode, DAGInNode, and DAGOutNode."""

    def __init__(self, wire=None, node_id=-1):
        self._wire = wire
        self._node_id = node_id

    def __lt__(self, other):
        return self._node_id < other._node_id

    def __gt__(self, other):
        return self._node_id > other._node_id

    def __str__(self):
        return str(id(self))


class DAGOpNode(DAGNode):
    """Object to represent an Instruction at a node in the DAGCircuit."""

    def __init__(self, op, qargs=None, cargs=None):
        super().__init__()
        self._type = "op"
        self.op = op
        self.qargs = qargs
        self.cargs = cargs
        self.sort_key = str(self.qargs)

    @property
    def name(self):
        """Operation name."""
        return str(self.op)

    def __repr__(self):
        return f"DAGOpNode(op={self.op}, qargs={self.qargs}, cargs={self.cargs})"

    __str__ = __repr__


class DAGInNode(DAGNode):
    """Object to represent an incoming wire node in the DAGCircuit."""

    def __init__(self, wire):
        super().__init__()
        self._type = "in"
        self.wire = wire
        self.sort_key = str([])

    def __repr__(self):
        return f"DAGInNode(wire={self.wire})"

    __str__ = __repr__


class DAGOutNode(DAGNode):
    """Object to represent an outgoing wire node in the DAGCircuit."""

    def __init__(self, wire):
        super().__init__()
        self._type = "out"
        self.wire = wire
        self.sort_key = str([])

    def __repr__(self):
        return f"DAGOutNode(wire={self.wire})"

    __str__ = __repr__



"""Self-increment operation  example."""

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import Measure, All
from qutrunk.circuit.ops import Classical
from qutrunk.circuit.ops import INC


def increment(num_qubits, init_value):
    # Create quantum circuit
    circuit = QCircuit()

    # Allocate quantum qubits
    qr = circuit.allocate(num_qubits)

    # Set initial amplitudes to classical state with init_value
    Classical(init_value) * qr

    # Apply quantum gates
    INC * qr

    # Measure all quantum qubits
    All(Measure) * qr

    # Run quantum circuit
    res = circuit.run()

    # Print measure result like:
    # 0b0001
    print(res.get_outcome())

    return circuit


if __name__ == "__main__":
    # Run locally
    circuit = increment(4, 0)

    # Dram quantum circuit
    circuit.draw()



"""Definition of some meta operator."""
from typing import Union, Optional
import numpy as np

from .basicgate import BasicGate
from qutrunk.circuit import QuBit, Qureg
from qutrunk.circuit.command import Command, CmdEx, Mat


# need to improve.
class All(BasicGate):
    """Meta operator, provides unified operation of multiple qubits.

    Args:
        gate: The gate will apply to all qubits.

    Example:
        .. code-block:: python

            All(H) * qureg
            All(Measure) * qureg
    """

    def __init__(self, gate):
        self.gate = gate

    def __or__(self, qureg):
        """Quantum logic gate operation.

        Args:
            qureg: The qureg(represent a set of qubit) to apply gate.

        Example:
            .. code-block:: python

                All(H) * qureg
                All(Measure) * qureg
        """
        for q in qureg:
            self.gate * q

    def __mul__(self, qureg):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement."""
        self.__or__(qureg)


class Power(BasicGate):
    """Power Gate.

    Args:
        power: The power to raise target gate to.
        gate: The target gate to raise.

    Example:
        .. code-block:: python

            Power(2, gate) * q[0]
    """

    def __init__(self, power, gate):
        self.power = power
        self.gate = gate

    def __or__(self, qubits: Union[QuBit, Qureg, tuple]):
        """Quantum logic gate operation."""
        if self.power < 0:
            raise ValueError("power should >= 0.")

        if not isinstance(qubits, (QuBit, Qureg, tuple)):
            raise TypeError("qubits should be type of QuBit, Qureg or tuple.")

        for _ in range(self.power):
            self.gate * qubits

    def __mul__(self, qubits: Union[QuBit, Qureg, tuple]):
        self.__or__(qubits)


class Matrix(BasicGate):
    """Custom matrix gate.

    Example:
            .. code-block:: python

                Matrix([[0.5, 0.5], [0.5, -0.5]]) * qr[0]  -- No controlled bit
                Matrix([[0.5, 0.5], [0.5, -0.5]], 1) * (qr[0], qr[1])  -- qr[0] is controlled bit
                Matrix([[0.5, 0.5], [0.5, -0.5]], 2) * (qr[0], qr[1], qr[2])  -- qr[0], qr[1] are controlled bits
    """

    def __init__(self, matrix, ctrl_cnt=0):
        super().__init__()
        self.matrix = matrix
        self.ctrl_cnt = ctrl_cnt

    def __str__(self):
        return "Matrix"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply X gate.

        Raises:
            NotImplementedError: If the argument is not a Qubit object.
        """
        if not isinstance(qubits, QuBit) and not all(
            isinstance(qubit, QuBit) for qubit in qubits
        ):
            raise TypeError("The argument must be Qubit object.")

        if (isinstance(qubits, QuBit) and self.ctrl_cnt > 0) or (
            not isinstance(qubits, QuBit) and (len(qubits) <= self.ctrl_cnt)
        ):
            raise ValueError("The parameter miss controlled or target qubit(s).")

        controls = (
            None if self.ctrl_cnt <= 0 else [q.index for q in qubits[0 : self.ctrl_cnt]]
        )
        targets = (
            [qubits.index]
            if isinstance(qubits, QuBit)
            else [q.index for q in qubits[self.ctrl_cnt :]]
        )

        if not self.check_matrix_format(len(targets)):
            raise ValueError(
                "The matrix is not in the right format by specified target(s)."
            )

        e = np.matrix(self.matrix)
        if self.is_inverse:
            if not self.is_unitary(e):
                raise ValueError(
                    "Only unitary matrices support invertible operations"
                )
            else:
                e = e.T.conjugate()

        cmd = Command(self, targets, controls, cmdex=CmdEx(mat=Mat()))
        cmd.cmdex.mat.reals = np.real(e).tolist()
        cmd.cmdex.mat.imags = np.imag(e).tolist()
        cmd.cmdex.mat.unitary = self.is_unitary(e)

        self.commit(qubits.circuit, cmd) if isinstance(qubits, QuBit) else self.commit(
            qubits[0].circuit, cmd
        )

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubit)

    def inv(self):
        """Apply inverse gate."""
        gate = Matrix(self.matrix, self.ctrl_cnt)
        gate.is_inverse = not self.is_inverse
        return gate

    def ctrl(self, ctrl_cnt=1):
        """Apply controlled gate.

        Args:
            ctrl_cnt: The number of control qubits, default: 1.
        """
        gate = Matrix(self.matrix, ctrl_cnt)
        gate.is_inverse = self.is_inverse
        return gate

    def check_matrix_format(self, numtargets):
        len_targets = 2**numtargets
        len_matrix = len(self.matrix)
        if len_targets != len_matrix:
            return False

        for row in range(len(self.matrix)):
            len_row = len(self.matrix[row])
            if len_targets != len_row:
                return False

        return True

    def is_unitary(self, mat):
        """Test a matrix is unitary or not.

         Example:
            .. code-block:: python

                m = [[1, 0], [0, 1]]
                m = np.matrix(m)
                print(is_unitary(m))
        """
        return np.allclose(np.eye(mat.shape[0]), mat.H * mat)


class Gate(BasicGate):
    """Definition of custom gate.

    Implement by composing some basic logic gates or define specific matrix.

    Example:
        .. code-block:: python

            @Gate
            def my_gate(a, b, c, d):
                return Gate() << (Matrix([[-0.5, 0.5], [0.5, 0.5]], 2).inv(), (a, b, c)) \
                    << (Matrix([[0.5, -0.5], [0.5, 0.5]]).ctrl().inv(), (a, c)) \
                    << (Matrix([[0.5, 0.5], [-0.5, 0.5]]), b)

            my_gate * (q[3], q[1], q[0], q[2])
    """

    def __init__(self, func: Optional[callable] = None):
        super().__init__()
        self.gates = []
        self.func = func

    def __lshift__(self, gate_define):
        if not isinstance(gate_define[0], BasicGate):
            raise TypeError("The first parameter is not a gate object.")

        if not isinstance(gate_define[1], QuBit) and not all(
            isinstance(qubit, QuBit) for qubit in gate_define[1]
        ):
            raise TypeError("The argument must be Qubit object.")

        self.gates.append({"gate": gate_define[0], "qubits": gate_define[1]})

        return self

    def __or__(self, qubits: Union[QuBit, tuple]):
        """Quantum logic gate operation."""
        if not isinstance(qubits, QuBit) and not all(
            isinstance(qubit, QuBit) for qubit in qubits
        ):
            raise TypeError("The argument must be Qubit object.")

        if isinstance(qubits, QuBit):
            custom_gate = self.func(qubits)
        else:
            custom_gate = self.func(*qubits)

        for c in custom_gate.gates:
            c["gate"] * c["qubits"]

    def __mul__(self, qubits: Union[QuBit, tuple]):
        self.__or__(qubits)



import os
import uuid

from thrift.protocol import TBinaryProtocol, TMultiplexedProtocol
from thrift.transport import TSocket, TTransport

from qutrunk.sim.qusprout.qusprout import QuSproutServer
from qutrunk.sim.qusprout.qusproutdata import ttypes as qusproutdata
from qutrunk.tools.function_time import timefn


class QuSproutApiServer:
    """
    The RPC client connect to QuSprout. Used by BackendQuSprout.

    Args:
        ip: ip address.
        port: port, default: 9090.
    """

    def __init__(self, ip="localhost", port=9090):
        socket = TSocket.TSocket(ip, port)
        self._transport = TTransport.TBufferedTransport(socket)
        protocol = TBinaryProtocol.TBinaryProtocol(self._transport)
        quest = TMultiplexedProtocol.TMultiplexedProtocol(protocol, "QuSproutServer")
        self._client = QuSproutServer.Client(quest)
        self._taskid = uuid.uuid4().hex
        try:
            self._transport.open()
        except Exception:
            print("QuSprout is not available!")
            os._exit(1)

    def close(self):
        self._transport.close()

    @timefn
    def init(self, qubits, density, exectype=0):
        """
        Initialize the quantum computing program at backend.

        Args:
            qubits: The number of qubits used in the entire circuit.
            state: Init state.
            value: Init Value.
            density: Enable noise model.
            exectype: SingleProcess: use single calculation node; \
                Mpi: parallel calculation using multiple nodes.
        """
        req = qusproutdata.InitQubitsReq(self._taskid, qubits, density, exectype)
        return self._client.initQubits(req)

    @timefn
    def send_circuit(self, circuit, final):
        """
        Args:
            circuit: Quantum circuit to send.
            final: True if quantum circuit finish, default False, \
            when final==True The backend program will release the computing resources.
        """
        req = qusproutdata.SendCircuitCmdReq(self._taskid, circuit, final)
        res = self._client.sendCircuitCmd(req)
        return res

    @timefn
    def run(self, shots):
        """Run quantum circuit.

        Args:
            shots: Circuit run times, for sampling, default: 1.

        Returns:
            result: The Result object contain circuit running outcome.
        """
        req = qusproutdata.RunCircuitReq(self._taskid, shots)
        res = self._client.run(req)
        return res.result

    @timefn
    def get_prob(self, index):
        """Get the probability of a state-vector at an index in the full state vector.

        Args:
            index: Index in state vector of probability amplitudes

        Returns:
            The probability of target index
        """
        req = qusproutdata.GetProbAmpReq(self._taskid, index)
        res = self._client.getProbAmp(req)
        return res.amp

    @timefn
    def get_probs(self, qubits):
        """Get all probabilities of circuit.

        Returns:
            An array contains all probabilities of circuit.
        """
        req = qusproutdata.GetProbOfAllOutcomReq(self._taskid, qubits)
        res = self._client.getProbOfAllOutcome(req)
        return res.pro_outcomes

    @timefn
    def get_statevector(self):
        """Get the current state vector of probability amplitudes for a set of qubits.

        Returns:
            Array contains all amplitudes of state vector.
        """
        req = qusproutdata.GetAllStateReq(self._taskid)
        res = self._client.getAllState(req)
        return res.all_state

    @timefn
    def cancel_cmd(self):
        """Cancel current job."""
        req = qusproutdata.CancelCmdReq(self._taskid)
        return self._client.cancelCmd(req)

    @timefn
    def get_expec_pauli_prod(self, pauli_prod_list):
        """Computes the expected value of a product of Pauli operators.

        Args:
            pauli_prod_list: A list contains the indices of the target qubits,\
                the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) to apply to the corresponding qubits.

        Returns:
            The expected value of a product of Pauli operators.
        """
        req = qusproutdata.GetExpecPauliProdReq(self._taskid, pauli_prod_list)
        res = self._client.getExpecPauliProd(req)
        return res.expect

    @timefn
    def get_expec_pauli_sum(self, oper_type_list, term_coeff_list):
        """
        Computes the expected value of a sum of products of Pauli operators.

        Args:
            oper_type_list: a list of the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) \
                of all Paulis involved in the products of terms. A Pauli must be specified \
                for each qubit in the register, in every term of the sum.
            term_coeff_list: the coefficients of each term in the sum of Pauli products.

        Returns:
            the expected value of a sum of products of Pauli operators.
        """
        req = qusproutdata.GetExpecPauliSumReq(
            self._taskid, oper_type_list, term_coeff_list
        )
        res = self._client.getExpecPauliSum(req)
        return res.expect
